import nyvalis

# from nyvalis.tools import log
# log.Env.init().stdout().file("log.txt", json=True)
# log.info("This is my log message!")

@nyvalis.command
def example(name: str):
    return f"{name}, Hello from Python!"


if __name__ == "__main__":
    nyvalis.Builder.default().invokers([example]).workers(2).run()


class NyvalisInvokerError(Exception):
    """Exception for Invoker Handler"""
    pass

class NyvalisInvokerAlreadyRegistered(Exception):
    """Exception in case command already is registered"""
    pass



import os
import uuid
import shutil
import requests
import subprocess
from typing import Dict
from pathlib import Path
from nyvalis.core import LIB_NAME
from nyvalis.bundler import utils
from nyvalis.tools import stdout, log
import xml.etree.ElementTree as xml_et

logger = log.get(log.LIB_NAME)
DEFAULT_WXS_PATH = os.path.join(os.path.dirname(__file__), "default.wxs")


def compile_installer(path: str, config: Dict):
    
    temp_path = Path(os.path.dirname(path)) / "temp"
    os.makedirs(temp_path, exist_ok=True)
    wix_path = get_wix_path(temp_path)

    wxs_path = None
    msi_obj = resolve_msi_bundler_obj(config)

    if msi_obj:
        wxs_path = msi_obj.get("path", None)

    temp_wxs = os.path.join(temp_path, "temp.wxs")
    if not wxs_path:
        logger.info("No custom wxs found, using default")
        logger.info("Copying wxs template to temp folder")
        shutil.copy2(DEFAULT_WXS_PATH, temp_wxs)
        wxs_path = temp_wxs
        resolve_default_wxs_file(temp_wxs, config, path)

    candle_exe = wix_path / "candle.exe"
    light_exe = wix_path / "light.exe"

    candle_cmd = [
        str(candle_exe),
        "-arch", utils.get_system_architecture(),
        "-out", os.path.join(temp_path, "output.wixobj"), 
        str(wxs_path)
    ]

    logger.info("Running candle.exe")
    candle_process = subprocess.run(candle_cmd, capture_output=True, text=True)
    if candle_process.returncode != 0:
        logger.error(candle_process.stdout)
        stdout.error(candle_process.stdout, exit=True)

    msi_dir = os.path.join(os.path.dirname(path), "bundles", "msi")
    os.makedirs(msi_dir, exist_ok=True)

    msi_name = f"{config['productName']}-{config["version"]}.msi"
    msi_output = os.path.join(msi_dir, msi_name)

    light_cmd = [
        str(light_exe),
        os.path.join(temp_path, "output.wixobj"),
        "-out", f'{msi_output}',
        "-ext", "WixUIExtension",
    ]

    logger.info("Running light.exe")
    light_process = subprocess.run(light_cmd, capture_output=True, text=True)
    if light_process.returncode != 0:
        logger.error(light_process.stdout)
        stdout.error(light_process.stdout, exit=True)

    stdout.info(f"msi installer created at: `{Path(msi_output).parent}`")
    shutil.rmtree(temp_path)

def resolve_default_wxs_file(path, config, exe_path):
    tree = xml_et.parse(path)
    root = tree.getroot()

    exe_path = str(Path(exe_path).resolve())

    logger.info("Updating values in temp.wxs")

    product_name = config["productName"]
    version = config["version"]
    manufacturer = LIB_NAME
    # icon_path = config["iconPath"]
    identifier = config["identifier"]
    guid = str(uuid.uuid5(uuid.NAMESPACE_DNS, identifier))

    ns = {"wix": "http://schemas.microsoft.com/wix/2006/wi"}

    product_node = root.find(".//wix:Product", ns)
    product_node.set("Name", product_name)
    product_node.set("Version", version)
    product_node.set("Manufacturer", manufacturer)
    product_node.set("UpgradeCode", guid)

    root.find(".//wix:Property[@Id='WIXUI_EXITDIALOGOPTIONALCHECKBOXTEXT']", ns).set("Value", f"Run {product_name}")
    root.find(".//wix:Directory[@Id='INSTALLFOLDER']", ns).set("Name", product_name)

    component = root.find(".//wix:Directory[@Id='INSTALLFOLDER']", ns)
    
    component.find(".//wix:RegistryValue", ns).set("Key", f"Software\\{manufacturer}\\{product_name}")
    component.find(".//wix:File", ns).set("Source", exe_path)


    shortcut_component = root.find(".//wix:Component[@Id='DesktopShortcutComponent']", ns)
    shortcut_component.find(".//wix:Shortcut", ns).set("Name", product_name)
    shortcut_component.find(".//wix:Shortcut", ns).set("Target", f"[INSTALLFOLDER]{product_name}.exe")
    shortcut_component.find(".//wix:RegistryValue", ns).set("Key", f"Software\\{manufacturer}\\{product_name}")
    
    logger.info("Saving values in temp.wxs")
    tree.write(path, encoding="utf-8", xml_declaration=True)


def download_wix_toolset(temp_path: Path):
    wix_url = "https://github.com/wixtoolset/wix3/releases/download/wix3141rtm/wix314-binaries.zip"
    
    logger.info("Downloading wix toolset")
    response = requests.get(wix_url, stream=True)
    zip_path = temp_path / "wix.zip"
    
    with open(zip_path, 'wb') as f:
        for chunk in response.iter_content(chunk_size=8192):
            f.write(chunk)
    

    wix_path = Path(temp_path) / ".wix"
    logger.info("Unpacking files")
    shutil.unpack_archive(zip_path, wix_path)
    return wix_path

def get_wix_path(temp_path: Path) -> Path:

    path = Path(os.environ.get("WIX", ""))

    logger.info("Searching for installed wix path")
    if path.exists() and (path / "candle.exe").exists() and (path / "light.exe").exists():
        logger.info("Wix path found")
        return path
    
    logger.info("Could not find wix")
    return download_wix_toolset(temp_path)


def resolve_msi_bundler_obj(config):
    bundler_obj = config.get("bundler", None)
    if bundler_obj:
        return  bundler_obj.get("msi", None)
    return bundler_obj



import json
import logging
import datetime as dt

class Colors:
    DEBUG = "\033[94m"
    INFO = "\033[92m"
    WARNING = "\033[93m"
    ERROR = "\033[91m"
    CRITICAL = "\033[95m"
    EXCEPTION = "\033[95m"
    RESET = "\033[0m"

RECORD_DEFAULTS = {
    "args",
    "asctime",
    "created",
    "exc_info",
    "exc_text",
    "filename",
    "funcName",
    "levelname",
    "levelno",
    "lineno",
    "module",
    "msecs",
    "message",
    "msg",
    "name",
    "pathname",
    "process",
    "processName",
    "relativeCreated",
    "stack_info",
    "thread",
    "threadName",
    "taskName",
}

DEFAULT_JSON_FMT = {
    "asctime": "timestamp",
    "name": "name",
    "module": "module",
    "funcName": "function",
    "levelname": "level",
    "message": "message",
    "pathname": "path",
    "lineno": "line"
}

class JSONFormatter(logging.Formatter):
    def __init__(self, json_fmt=None, iso: bool = False):
        super().__init__()
        self.iso = iso
        self.json_fmt = json_fmt or DEFAULT_JSON_FMT


    def format(self, record):
        log_record = {}

        for k, v in self.json_fmt.items():
            if k == "asctime":
                log_record[v] = self.formatTime(record, self.datefmt)
                if self.iso:
                    log_record[v] = format_to_iso(record.created)
            elif k == "message":
                log_record[v] = record.getMessage()
            else:
                log_record[v] = getattr(record, k, None)

        if record.exc_info is not None:
            log_record["exc_info"] = self.formatException(record.exc_info)

        if record.stack_info is not None:
            log_record["stack_info"] = self.formatStack(record.stack_info)

        for key, val in record.__dict__.items():
            if key not in RECORD_DEFAULTS:
                log_record[key] = val

        return json.dumps(log_record)

class CustomFormatter(logging.Formatter):
    def __init__(self, fmt=None, datefmt=None, colored: bool = False, iso: bool = False):
        super().__init__(fmt, datefmt)
        self.colored = colored
        self.iso = iso

    def format(self, record):
        asctime = format_to_iso(record.created)

        message = super().format(record)
        
        if self.colored:
            levelname = record.levelname
            colored_levelname = f"{Colors.__dict__.get(levelname, '')}{levelname}{Colors.RESET}"
            message = message.replace(levelname, colored_levelname, 1)

        if self.iso:
            message = message.replace(record.asctime, asctime, 1)
        
        return message
    

def format_to_iso(timestamp: float):
    local_timezone = dt.datetime.now().astimezone().tzinfo

    return dt.datetime.fromtimestamp(timestamp, tz=local_timezone) \
        .replace(microsecond=int((timestamp % 1000) * 1000)) \
        .isoformat(timespec='milliseconds')


import json
from .models import Command
from typing import Callable, Dict
from nyvalis.tools import stdout, log

class Invoker:
    commands: Dict[str, Command] = {}
    logger = log.get(log.LIB_NAME)

    @classmethod
    def register(cls, name: str, func: Callable):
        if cls.commands.get(name):
            cls.logger.critical(f"Command with same name was found: `{name}`")
            stdout.error(f"Command with same name was found: `{name}`", exit=True)
        
        cls.logger.info(f"Command `{name}` registered")
        cls.commands[name] = Command(name=func.__name__, allowed=False, func=func)

    @classmethod
    def process(cls, command: Command, params):
        try:
            result = command.func(**params)
            response = json.dumps({"data": str(result), "error": None})
            return response
        except Exception as err:
            cls.logger.error(err)
            response = json.dumps({"data": None, "error": str(err)})
            return response

def command(func):
    Invoker.commands[func.__name__] = Command(name=func.__name__, allowed=False, func=func)
    return func


import os
import sys
import json
import shutil
import subprocess
import questionary
import importlib.util
from typing import List
from pathlib import Path
import importlib.metadata
from nyvalis.tools import stdout, log
from nyvalis.core import CONF_FILENAME, LIB_NAME
from nyvalis.cli.commands.models import PackageManager, TemplateObj

logger = log.get(log.LIB_NAME)

def prompt_text(title: str, default):
    value = questionary.text(title, default=default).ask()

    if value is None:
        logger.info(f"{title}: Cancelled by user.")
        stdout.info("Operation cancelled by user", exit=True)
    
    return value


def prompt_select(title: str, choices: List[questionary.Choice]):
    value = questionary.select(
        title,
        choices=choices
    ).ask()

    if value is None:
        logger.info(f"{title}: Cancelled by user.")
        stdout.info("Operation cancelled by user", exit=True)
    
    return value

def prompt_checkbox(title: str, choices: List[questionary.Choice]):
    value = questionary.checkbox(
        title,
        choices=choices
    ).ask()

    if value is None:
        logger.info(f"{title}: Cancelled by user.")
        stdout.info("Operation cancelled by user", exit=True)
    
    return value


def resolve_conf_file(manager: PackageManager, template: TemplateObj, path: str):
    dst_json = os.path.join(path, "src-python", CONF_FILENAME)
    
    with open(dst_json, 'r') as file:
        config = json.load(file)

    name = "My App"
    if path != ".":
        name = path

    config["productName"] = name
    config["version"] = "0.1.0"
    config["identifier"] = f"com.{name}.app"
    config["window"]["title"] = name

    config["build"] = template.build_field.to_dict()

    for key in config["build"]:
        config["build"][key] = config["build"][key].replace(r"{{ manager }}", f"{manager.execute}")
    
    with open(dst_json, 'w') as file:
        json.dump(config, file, indent=4)

    log.info("Writing configuration file")

def scaffold_python(manager: PackageManager, path: str, template: TemplateObj):
    PYTHON_TEMPLATE_DIR = Path(os.path.dirname(os.path.dirname(os.path.abspath(__file__)))) / "templates" / "python"

    log.info("Scaffolding python")
    src_path = Path(path).resolve() / "src-python"
    shutil.copytree(PYTHON_TEMPLATE_DIR, src_path, dirs_exist_ok=True)

    resolve_conf_file(manager, template, path)
    log.info("Python template done")


def is_cli_installed(name: str, arg="-v") -> bool:
    try:
        subprocess.run([name, arg], capture_output=True, text=True, check=True, shell=True)
        return True
    except subprocess.CalledProcessError:
        return False
    except FileNotFoundError:
        return False


def create_git_repo(path: str):
    try:
        subprocess.run(['git', 'init'], cwd=path, check=True)
        log.info("Git repository initialized in output directory.")
        stdout.info("Git repository initialized in output directory.")
    except subprocess.CalledProcessError as err:
        log.error(f"Failed to initialize Git: {err}")
        stdout.error(f"Failed to initialize Git: {err}", exit=True)

def write_requirements(path: str):
    try:
        with open(f"{path}/requirements.txt", "w") as file:
            file.write(f"{LIB_NAME}=={importlib.metadata.version(LIB_NAME)}")
        log.info("Sucessfully created requirements.txt")
    except Exception as err:
        log.error(f"Failed to write requirements.txt: {err}")
        stdout.error(f"Failed to write requirements.txt: {err}", exit=True)
    

def setup_venv(path: str):
    stdout.info("Setting up venv...")
    try:
        subprocess.run([sys.executable, "-m", "venv", ".venv"], cwd=path, check=True)
        log.info("Venv created")
        stdout.info("Venv created")
    except Exception as err:
        log.error(f"Failed to set up venv: {err}")
        stdout.error(f"Failed to set up venv: {err}")
    

def run_pip_install(path: str, venv: bool):
    stdout.info("Running pip install")
    try:
        if venv:
            pip_command = os.path.join(path, ".venv", "Scripts", "pip")
        else:
            pip_command = "pip"
        subprocess.run([pip_command, "install", "-r", "requirements.txt"], cwd=path, check=True)
        log.info("Pip dependencies installed successfully.")
        stdout.info("Pip dependencies installed successfully.")

    except Exception as err:
        log.error(f"Failed to install pip dependencies: {err}")
        stdout.error(f"Failed to install pip dependencies: {err}")

def run_npm_install(path: str, manager: PackageManager):
    stdout.info(f"Running {manager.cli} install")
    subprocess.run([manager.cli, "install"], cwd=path, check=True, shell=True)
    log.info("Node dependencies installed successfully.")
    stdout.info("Node dependencies installed successfully.")

def get_package_path(package_name):
    package_spec = importlib.util.find_spec(package_name)
    return os.path.dirname(package_spec.origin)

def is_venv():
    return (hasattr(sys, 'real_prefix') or
            (hasattr(sys, 'base_prefix') and sys.base_prefix != sys.prefix))


import sys
import logging
from nyvalis.tools import log

logger = log.get(log.LIB_NAME)

if not logger.hasHandlers():
    logger.setLevel(logging.NOTSET)
    handler = logging.StreamHandler()
    handler.setFormatter(log.handlers.CustomFormatter(log.DEFAULT_FMT, log.DEFAULT_DATEFMT, colored=True, iso=True))
    logger.addHandler(handler)
    logger.propagate = False

if "-v" in sys.argv:
    logger.setLevel(logging.DEBUG)


from nyvalis.core.builder import Builder
from nyvalis.core.invoker import command



import os
import clr
import pythonnet
from pathlib import Path

ROOT_PATH = Path(os.path.dirname(__file__)).resolve().parents[1]
DLL_PATH = os.path.join(os.path.dirname(__file__), "dlls")
LIB_NAME = "nyvalis"
CONF_FILENAME = f"{LIB_NAME}.conf.json"

pythonnet.load("coreclr")
clr.AddReference("System")
clr.AddReference("System.Windows.Forms")

clr.AddReference(os.path.join(DLL_PATH, "Microsoft.Web.WebView2.Core.dll"))
clr.AddReference(os.path.join(DLL_PATH, "Microsoft.Web.WebView2.WinForms.dll"))



import platform

def get_system_architecture() -> str:
    arch = platform.machine().lower()
    if arch in ['x86_64', 'amd64']:
        return 'x64'
    elif arch in ['x86', 'i686']:
        return 'x86'
    elif arch == 'arm64':
        return 'arm64'
    else:
        return 'x64'


from System.Reflection import BindingFlags, Assembly
from System.Windows.Forms import OpenFileDialog, DialogResult, MessageBox, MessageBoxButtons
from System import UInt32, Object, Array


from typing import List
from dataclasses import dataclass

@dataclass
class FileFilter:
    text: str
    extension: List[str]

    def __str__(self):
        ext_str = ";".join(f"*{ext}" for ext in self.extension)
        return f"{self.text} ({ext_str})|{ext_str}"


class Dialog:
    __FOLDERS_FILTER = 'Folders'
    __FLAGS = BindingFlags.Instance | BindingFlags.Public | BindingFlags.NonPublic
    __winforms_assembly = Assembly.LoadWithPartialName('System.Windows.Forms')

    __i_file_dialog_type = __winforms_assembly.GetType('System.Windows.Forms.FileDialogNative+IFileDialog')
    __open_file_dialog_type = __winforms_assembly.GetType('System.Windows.Forms.OpenFileDialog')
    __file_dialog_type = __winforms_assembly.GetType('System.Windows.Forms.FileDialog')
    __create_vista_dialog = __open_file_dialog_type.GetMethod('CreateVistaDialog', __FLAGS)
    __on_before_vista_dialog = __open_file_dialog_type.GetMethod('OnBeforeVistaDialog', __FLAGS)
    __get_options = __file_dialog_type.GetMethod('GetOptions', __FLAGS)
    __set_options = __i_file_dialog_type.GetMethod('SetOptions', __FLAGS)
    __fos_pick_folders = __winforms_assembly.GetType('System.Windows.Forms.FileDialogNative+FOS')\
        .GetField('FOS_PICKFOLDERS').GetValue(None)

    __vista_dialog_events_constructor = __winforms_assembly.GetType(
        'System.Windows.Forms.FileDialog+VistaDialogEvents'
    ).GetConstructor(__FLAGS, None, [__file_dialog_type], [])
    __advise_method = __i_file_dialog_type.GetMethod('Advise')
    __show_method = __i_file_dialog_type.GetMethod('Show')

    @staticmethod
    def open(title: str, directory: bool = False, multiple: bool = False, path=None, filters: List[FileFilter]=[]):
        dialog = OpenFileDialog()
        dialog.Title = title
        dialog.Multiselect = multiple
        dialog.InitialDirectory = path

        if directory:
            dialog.Filter = Dialog.__FOLDERS_FILTER
            dialog.CheckFileExists = False
            dialog.DereferenceLinks = True
            dialog.AddExtension = False
            dialog.RestoreDirectory = True

            i_file_dialog = Dialog.__create_vista_dialog.Invoke(dialog, [])
            Dialog.__on_before_vista_dialog.Invoke(dialog, [i_file_dialog])
            options = Dialog.__get_options.Invoke(dialog, [])
            options = options.op_BitwiseOr(Dialog.__fos_pick_folders)
            Dialog.__set_options.Invoke(i_file_dialog, [options])

            advise_params = Array.CreateInstance(Object, 2)
            advise_params[0] = Dialog.__vista_dialog_events_constructor.Invoke([dialog])
            advise_params[1] = UInt32(0)
            Dialog.__advise_method.Invoke(i_file_dialog, advise_params)

            result = Dialog.__show_method.Invoke(i_file_dialog, [None])
            return list(dialog.FileNames) if result == 0 else None

        filter_strings = [str(f) for f in filters]
        filter_strings.append("All Files (*.*)|*.*")
        filter = "|".join(filter_strings)

        dialog.Filter = filter if filters else "All Files|*.*" 
        dialog.RestoreDirectory = True
        result = dialog.ShowDialog()
        return list(dialog.FileNames) if result == DialogResult.OK else None

    @staticmethod
    def confirmation(title: str, message: str) -> bool:
        result = MessageBox.Show(message, title, MessageBoxButtons.OKCancel)
        return result == DialogResult.OK
    
open = Dialog.open
confirmation = Dialog.confirmation



from nyvalis.tools import log
from nyvalis.core.config import Config
from nyvalis.core import webview, ROOT_PATH

from System import Func, Type
from System.Drawing import Size, Icon
from System.Windows.Forms import Form, Application, FormBorderStyle


class WinForms(Form):
    logger = log.get(log.LIB_NAME)

    def __init__(self, config: Config, workers: int):
        super().__init__()
        self.uid = config.identifier

        self.Text = config.window.title
        self.Size = Size(config.window.width, config.window.height)
        self.MinimumSize = Size(config.window.min_width, config.window.min_height)

        self.logger.info(f"WinForms Window UUID: {config.identifier}")
        self.logger.info(f"WinForms Window Text: {config.window.title}")
        self.logger.info(f"WinForms Window Size: {config.window.width}x{config.window.height}")
        self.logger.info(f"WinForms Window MinimumSize: {config.window.min_width}x{config.window.min_height}")

        self.logger.info("WinForms successfully configurated")

        if config.icon:

            if not "__compiled__" in globals():
                self.Icon = Icon(config.icon)
            else:
                icon_path = ROOT_PATH / "icon.ico"
                self.Icon = Icon(str(icon_path))
            self.logger.info(f"WinForms Icon: {config.icon}")


        if not config.window.resizable:
            self.FormBorderStyle = FormBorderStyle.FixedSingle
        self.logger.info(f"WinForms Window Resizable: {config.window.resizable}")
        
        self.browser = webview.Webview(self, config, workers)
        self.FormClosed += self.on_exit

    def on_exit(self, *_):
        self.logger.info("WinForms Exit requested")
        self.Invoke(Func[Type](self.Hide))
        self.browser.on_exit()
        Application.Exit()

        self.logger.info("WinForms process terminated")

def initialize(config: Config, workers: int):
    """Starts the application"""
    logger = log.get(log.LIB_NAME)

    form = WinForms(config, workers)
    form.Show()

    app = Application
    logger.info("Application started")
    app.Run()
